#ifndef BITBOARD_H
#define BITBOARD_H

#include "assert.h"
#include "itype.h"
#include "sbr.h"

typedef uint64 BITBOARD;

extern const BITBOARD board_empty;
extern const BITBOARD board_all;
extern const BITBOARD board_minus_a;
extern const BITBOARD board_minus_h;
extern const BITBOARD board_centre;
extern const BITBOARD board_xcentre;
extern const BITBOARD board_kwing;
extern const BITBOARD board_qwing;
extern const BITBOARD board_rank1;
extern const BITBOARD board_rank2;
extern const BITBOARD board_rank3;
extern const BITBOARD board_rank4;
extern const BITBOARD board_rank5;
extern const BITBOARD board_rank6;
extern const BITBOARD board_rank7;
extern const BITBOARD board_rank8;

extern void printf_bitboard(BITBOARD b);

/*******************************************/
/* Bitboard manipulation routines (inline) */
/*******************************************/

/* Set a bit on the bitboard */
extern inline void set_bitboard(BITBOARD *board, uint8 square)
{
   assert(board);
   (*board) |= ((uint64)1)<<square;
}

/* Toggles a square on a bitboard */
extern inline void toggle_bitboard(BITBOARD *board, uint8 square)
{
   assert(board);
   (*board) ^= ((uint64)1)<<square;
}

/* Clears a square on the bitboard */
extern inline void unset_bitboard(BITBOARD *board, uint8 square)
{
   assert(board);
   (*board) &= ((uint64)0xFFFFFFFFFFFFFFFFll) - ((uint64)1<<square);
}

/* Returns true if a square on the bitboard is set */
extern inline int get_bitboard(BITBOARD *board, uint8 square)
{
   assert(board);
   return ((*board) & ((uint64)1)<<square)!=0;
}

/* ANDs two bitboards */
extern inline BITBOARD and_bitboards(const BITBOARD b1, const BITBOARD b2)
{
   return b1&b2;
}

/* XORs two bitboards */
extern inline BITBOARD xor_bitboards(const BITBOARD b1, const BITBOARD b2)
{
   return b1^b2;
}

/* ORs two bitboards */
extern inline BITBOARD or_bitboards(const BITBOARD b1, const BITBOARD b2)
{
   return b1|b2;
}

/* Inverts a bitboard */
extern inline BITBOARD invert_bitboard(const BITBOARD b)
{
   return ~b;
}

/* Retrieve one row from a bitboard */
extern inline uint8 get_bitboard_row(const BITBOARD b, const int row)
{
   return b >> (row*8);
}

/* Expand one row into a full bitboard */
extern inline BITBOARD make_bitboard_row(const uint8 data, const uint8 row_nr)
{
   return ((uint64)data) << (row_nr*8);
}

/* Shift a bitboard to the right (that is, a->b, b->c etc.) by 1 file */
/*  and discards what gets shifted in on the left */
/* Note that shifting the board to the right implies shifting the binary */
/*  values to the left, since files to the right have a higher index */
extern inline BITBOARD shift_bitboard_file_right(const BITBOARD b)
{
   return and_bitboards(board_minus_a, (b<<1));
}

/* Shift a bitboard to the left (that is, b->a, c->b etc.) by 1 file */
/*  and discards what gets shifted in on the right */
/* Note that shifting the board to the left implies shifting the binary */
/*  values to the right, since files to the left have a lower index */
extern inline BITBOARD shift_bitboard_file_left(const BITBOARD b)
{
   return and_bitboards(board_minus_h, (b>>1));
}

/* Shift a bitboard up one row (that is, 1->2, 2->3 etc */
extern inline BITBOARD shift_bitboard_row_up(const BITBOARD b)
{
   return b<<8;
}

/* Shift a bitboard down one row (that is, 1->2, 2->3 etc */
extern inline BITBOARD shift_bitboard_row_down(const BITBOARD b)
{
   return b>>8;
}

/* Combinations of the above */
extern inline BITBOARD shift_bitboard_right_up(const BITBOARD b)
{
   return shift_bitboard_row_up(shift_bitboard_file_right(b));
}

extern inline BITBOARD shift_bitboard_right_down(const BITBOARD b)
{
   return shift_bitboard_row_down(shift_bitboard_file_right(b));
}

extern inline BITBOARD shift_bitboard_left_up(const BITBOARD b)
{
   return shift_bitboard_row_up(shift_bitboard_file_left(b));
}

extern inline BITBOARD shift_bitboard_left_down(const BITBOARD b)
{
   return shift_bitboard_row_down(shift_bitboard_file_left(b));
}
#endif
