#ifndef GAME_H
#define GAME_H

#include "assert.h"

#include "linked.h"
#include "itype.h"
#include "board.h"
#include "move.h"
#include "movegen.h"

#define MAX_VARIANT_LENGTH 32
#define MAX_KILLER_DEPTH   16

typedef struct VARIANT {
   int movecount;                   /* Number of moves in variant */
   MOVE move[MAX_VARIANT_LENGTH];   /* Move sequence */
} VARIANT;

typedef struct GAME_POSITION_DATA {
   MOVE move;
#ifdef ENPASSANT_MOVE               /* Defined for chess, but not Xiangqi */
   int ep_square;
#endif
} GAME_POSITION_DATA;

typedef struct GAMESTATE {
   BOARD board;
   /* The position state: the player to move, the possible moves, */
   /*  material balance and position evaluation */
   MOVE_GEN_STRUCT mgl;
   int player;
   int material[2];
   int eval;
   
   /* List of killers starting at the current node (player/depth/number) */
   MOVE killer[2][MAX_KILLER_DEPTH][2];
   
   /* The principle variation */
   VARIANT pv;
   
   /* Position hash keys */
   int hash1;
   int hash2;

   /* The gamestate: moves played */
   LINK *movelist;
   LINK *movelist_tail;
   int moves_played;
} GAMESTATE;

extern void playgame_init(void);
extern void playgame_shutdown(void);

extern void start_newgame(GAMESTATE *game);
extern void end_game(GAMESTATE *game);

extern void count_material(GAMESTATE *game);

extern void forward_move(BOARD *board, const MOVE move);
extern void backward_move(BOARD *board, const MOVE move);

extern MOVE get_move_history(GAMESTATE *game, int player, int depth);

extern void playmove(GAMESTATE *game, MOVE move);
extern void takeback(GAMESTATE *game);

extern char *get_move_string_algebraic(const BOARD *b, const MOVE m, char *s);

/* Accessing the killer table */
extern inline void store_killer(GAMESTATE *game, MOVE move, int depth)
{
   int player;
   
   assert(depth>0);
   assert(depth<MAX_KILLER_DEPTH);
   
   player = get_move_player(move)>>7;
   
   /* Do nothing if the move is already the first killer */
   if (moves_are_equal(game->killer[player][depth][0], move))
      return;

   /* Store as first killer, demote old one to second killer */
   game->killer[player][depth][1] = game->killer[player][depth][0];
   game->killer[player][depth][0] = move;
}

/* Returns 1 if a move is a killer move, 0 if it is not. */
extern inline int is_killer(GAMESTATE *game, MOVE move, int depth)
{
   int player;
   
   assert(depth>0);
   assert(depth<MAX_KILLER_DEPTH);
   
   player = get_move_player(move)>>7;
   
   /* Do nothing if the move is already the first killer */
   if (moves_are_equal(game->killer[player][depth][0], move) || moves_are_equal(game->killer[player][depth][1], move))
      return 1;

   return 0;
}
#endif
