#ifndef MOVE_H
#define MOVE_H

#include "assert.h"
#include "itype.h"
#include "pieces.h"

#define MAXMOVES 128

/* Define structure to hold a chess move */
/* Layout: piece ID, from, to, ep-square */
/* Each fit in one char, four of which make up an int */
typedef union MOVE {
   struct {
      uint8 piece;      /* Piece on source square */
      uint8 from;       /* Origin square */
      uint8 to;         /* Destination square */
      uint8 taken;      /* Piece on destination square */
   } field;
   uint32 m;
} MOVE;

extern MOVE null_move;

/* Inlinable functions */
/* Get the player that makes the move */
extern inline uint8 get_move_player(const MOVE move)
{
   return move.field.piece&BLACK;
}

/* Get the piece that moves */
extern inline uint8 get_move_piece(const MOVE move)
{
   return move.field.piece;
}

/* Get the origin square */
extern inline uint8 get_origin(const MOVE move)
{
   return move.field.from;
}

/* Get the destination square */
extern inline uint8 get_destination(const MOVE move)
{
   return move.field.to;
}

/* Get the captured piece (if any) */
extern inline uint8 get_captured_piece(const MOVE move)
{
   return move.field.taken;
}

extern inline void encode_move(MOVE *move, const uint8 piece, const uint8 from, const uint8 to, const uint8 taken)
{
   assert(move);
   move->field.piece = piece;
   move->field.from = from;
   move->field.to = to;
   move->field.taken = taken;
}

/* Compare two moves */
extern inline int moves_are_equal(const MOVE m1, const MOVE m2)
{
   return m1.m == m2.m;
}

/****************/
/*  Move types  */
/****************/
/* Returns true if a move is a promotion move */
#ifdef PROMOTION_MOVE               /* Defined for chess, but not Xiangqi */
extern inline int is_promotion(const MOVE move)
{
   return (move.field.piece & PAWN) && (move.field.piece & (QUEEN|ROOK|BISHOP|KNIGHT));
}
#endif

/* Returns TRUE if the move is an en-passant move */
/* Technically, it incorrectly returns false when the ep-square  A1, which */
/*  is fortunately impossible */
#ifdef ENPASSANT_MOVE               /* Defined for chess, but not Xiangqi */
extern inline int is_enpassant(const MOVE move)
{
   return (move.field.piece & ENPASSANT_MOVE) == ENPASSANT_MOVE;
}
#endif

/* Returns TRUE if the move is a double pawn move */
#ifdef ENPASSANT_MOVE               /* Defined for chess, but not Xiangqi */
extern inline int is_doublepawn(const MOVE move)
{
   return (move.field.piece & PAWN) && (abs(move.field.from-move.field.to)==16);
}
#endif

/* returns TRUE is a move is a castling move */
#ifdef CASTLE_MOVE                  /* Defined for chess, but not Xiangqi */
extern inline int is_castling(const MOVE move)
{
   return (move.field.piece & CASTLE_MOVE) == CASTLE_MOVE;
}
#endif

/* returns TRUE is a move is a capture move */
extern inline int is_capture(const MOVE move)
{
   return move.field.taken;
}

/* Get promotion piece */
#ifdef PROMOTION_MOVE               /* Defined for chess, but not Xiangqi */
extern inline int get_promotion_piece(const MOVE move)
{
   assert(is_promotion(move));
   
   return move.field.piece & (QUEEN|ROOK|BISHOP|KNIGHT);
}
#endif

/*************************************************/
/*  Get en-passant square, captured square, etc. */
/*************************************************/
/* Get the en-passant square, ie, the square that held the original pawn */
/* For white, the is one row below the destination square, for black it is */
/*  one row above the destination square. */
#ifdef ENPASSANT_MOVE               /* Defined for chess, but not Xiangqi */
extern inline int get_ep_square(const MOVE move)
{
   assert(is_enpassant(move));
   /* If the piece colour is black, then */
   /*  ((move.field.piece & BLACK) >> 3) - 8 = 16-8 = 8, ie, one row above */
   /* If the piece colour is white, then */
   /*  ((move.field.piece & BLACK) >> 3) - 8 = 0-8 = -8, ie, one row below */
   return move.field.to + ( ((move.field.piece & BLACK) >> 3) - 8 );
}
#endif

/* Get the square the captured piece was on */
/* This is the destination square for ordinary moves and the ep square */
/*  for en-passant moves */
extern inline int get_captured_square(const MOVE move)
{
#ifdef ENPASSANT_MOVE               /* Defined for chess, but not Xiangqi */
   if (is_enpassant(move))
      return get_ep_square(move);
   else
#endif
      return move.field.to;
}

/* Get the Rook origin square in a castling move */
#ifdef CASTLE_MOVE                  /* Defined for chess, but not Xiangqi */
extern inline int get_rook_origin(const MOVE move)
{
   assert(is_castling(move));
   
   if (move.field.to > move.field.from)   /* Kingside */
      return move.field.from + 3;
   else                                   /* Queenside */
      return move.field.from - 4;
}
#endif

/* Get the Rook destination square in a castling move */
#ifdef CASTLE_MOVE                  /* Defined for chess, but not Xiangqi */
extern inline int get_rook_destination(const MOVE move)
{
   assert(is_castling(move));
   
   if (move.field.to > move.field.from)   /* Kingside */
      return move.field.from + 1;
   else                                   /* Queenside */
      return move.field.from - 1;
}
#endif
#endif
