#include "compplay.h"
#include "minimax.h"
#include "alphabet.h"
#include "mtdf.h"
#include "genrand.h"
#include "evaluate.h"
#include "hashtab.h"
#include "movegen.h"
#include "square.h"
#include "game.h"

#ifdef USE_PTHREADS
   #include <pthread.h>
#endif

#define printf dummy_printf

#if defined DJGPP && !defined printf
#define printf dummy_printf
#endif

static void dummy_printf(char *s, ...)
{
   return;
}

int depth_reached;

#ifdef USE_PTHREADS
static pthread_mutex_t computer_think_lock = PTHREAD_MUTEX_INITIALIZER;
#define LOCK()    pthread_mutex_lock(&computer_think_lock)
#define UNLOCK()  pthread_mutex_unlock(&computer_think_lock)
#else
#define LOCK()
#define UNLOCK()
#endif

/* Let the computer think and search for a move in the game */
/* Searches up to max_depth in the brute-force search */
/* Aborts when the volatile abort_evaluation variable from alphabeta */
/*  becomes non-zero and plays the best result found upto that point. */
void computer_play(GAMESTATE *game, const int max_depth)
{
   MOVE_GEN_STRUCT movelist;
   MOVE best_move;
   MOVE move;
   MOVE *valid_moves;
   int alpha, beta;
   int hash1, hash2;
   int lower, upper, h;
   int num_valid_moves;
   int best_value;
   int value;
   int plies;
   int best_move_nr = 0;
   int n;

   LOCK();

   assert(max_depth);

   prune_count = 0;
   evaluate_count = 0;

   alpha = -CHECKMATE;
   beta = CHECKMATE;

   /* Look for position in the hashtable */
   h = 0;
   hash1 = game->hash1;
   hash2 = game->hash2;

   if (retrieve_hashtable(hash1, hash2, &lower, &upper, &h, &best_move)) {
      if (lower>alpha)
         alpha = lower;

      if (upper<beta)
         beta  = upper;
   }

   /* Get a list of valid moves in this position */
   move_gen_start(&game->board, game->player, &movelist);

   /* Begin looking at each move in turn - if there are valid ones ofcourse */
   valid_moves = malloc(movelist.total.nummoves * sizeof *valid_moves);
   num_valid_moves = 0;
   while (get_next_move(&movelist, &move)) {
      playmove(game, move);
      if (!check(&game->board, game->player^BLACK)) {
         /* But best move from previous search in front */
         if (moves_are_equal(best_move, move)) {
            valid_moves[num_valid_moves++] = valid_moves[0];
            valid_moves[0] = move;
         } else {
            valid_moves[num_valid_moves++] = move;
         }
      }
      takeback(game);
   }


   depth_reached = 0;
   abort_evaluation = 0;

   hash_table_hits = 0;
   hash_table_misses = 0;
   hash_table_collissions = 0;

   value = best_value = -CHECKMATE;
   movelist.total.cmove = 0;

   best_move_nr = 0;
   best_move = valid_moves[0];

   if (num_valid_moves>1) {
      for (plies = 1; plies<=max_depth; plies++) {
         if (plies>1) {
            /* Adjust alpha/beta window for next iteration (we expect the value */
            /*  won't differ much from the current value) */
            beta = best_value + 100;
            alpha = best_value - 100;
         }

         for(n=0; n<num_valid_moves; n++) {
            move = valid_moves[n];

            playmove(game, move);
            /* Search principle variation one ply deeper than the rest */
            if (n==0)
               value = -alphabeta(game, -beta, -alpha, plies, &game->pv);
            else
               value = -alphabeta(game, -beta, -alpha, plies-1, &game->pv);
            takeback(game);

            /* Break out of evaluation if requested */
            if (abort_evaluation)
               break;

            /* When doing researches (in case of fail high or fail low), */
            /*  we need to open up the window to both sides slightly to  */
            /*  counter search instabilities. */
            if (n==0) {                   /* First move (best so-far) */
               /* Make sure value is within the search window */
               if (value<alpha) {         /* Best move fails low (ai!) */
                  alpha = -CHECKMATE;
                  beta = value;

                  /* Search again with the window opened below */
                  playmove(game, move);
                  value = -alphabeta(game, -beta, -alpha, plies, &game->pv);
                  takeback(game);
               } else if (value>=beta) {  /* Best move fails high */
                  alpha = value;
                  beta = CHECKMATE;

                  playmove(game, move);
                  value = -alphabeta(game, -beta, -alpha, plies, &game->pv);
                  takeback(game);
               }

               if (abort_evaluation)
                  break;
               if (value>best_value) {
                  best_value = value;
               }

               /* Just in case a next move is better, continue with minimal */
               /*  window search */
               alpha = value;
               beta = alpha+1;
            } else {
               if (value>=beta) {      /* Fail high - new best move found */
                  best_value = alpha;
                  alpha = value-100;
                  beta = CHECKMATE;

                  playmove(game, move);
                  value = -alphabeta(game, -beta, -alpha, plies, &game->pv);
                  takeback(game);

                  if (abort_evaluation)
                     break;

                  /* Make sure it is also better with the window opened */
                  if (value>best_value) {
                     /* Remember new best move */
                     best_value = value;
                     best_move = move;
                     best_move_nr = n;
                  }
                  alpha = value;
                  beta = alpha+1;
               }
            }
            printf ("\t[%d] %c%c - %c%c [%c%c - %c%c], score %d/%d (%d, %d)\n",
                    plies,
                    'a'+unpack_column(move.field.from),
                    '1'+unpack_row(move.field.from),
                    'a'+unpack_column(move.field.to),
                    '1'+unpack_row(move.field.to),
                    'a'+unpack_column(best_move.field.from),
                    '1'+unpack_row(best_move.field.from),
                    'a'+unpack_column(best_move.field.to),
                    '1'+unpack_row(best_move.field.to),
                    value, best_value, alpha, beta);
         }

         /* Break out of evaluation if requested */
         if (abort_evaluation)
            break;

         /* Put best move in the front */
         best_move = valid_moves[best_move_nr];
         for (n=best_move_nr; n>0; n--)
            valid_moves[n] = valid_moves[n-1];
         valid_moves[0] = best_move;
         best_move_nr = 0;

         game->eval = -alpha;

         depth_reached = plies;

         printf ("Best move to %d ply: %c%c - %c%c, score %d/%d\n", plies,
                 'a'+unpack_column(best_move.field.from),
                 '1'+unpack_row(best_move.field.from),
                 'a'+unpack_column(best_move.field.to),
                 '1'+unpack_row(best_move.field.to),
                 game->eval, best_value);

         printf("%d table hits, %d misses, %d collisions\n", hash_table_hits, hash_table_misses, hash_table_collissions);
         
         store_hashtable(hash1, hash2, best_value, best_value, depth_reached, best_move);

         /* Break if mate is inevitable */
         if (game->eval == CHECKMATE-plies+1)
            break;
         if (game->eval == -CHECKMATE+plies)
            break;
      }
   }
   printf ("Move: %c%c - %c%c, score %d\n\n",
           'a'+unpack_column(best_move.field.from),
           '1'+unpack_row(best_move.field.from),
           'a'+unpack_column(best_move.field.to),
           '1'+unpack_row(best_move.field.to),
            game->eval);

   if (num_valid_moves) {
      playmove(game, best_move);
   }

   free(valid_moves);
   
   //extract_principle_variation(game);

   UNLOCK();
}

void extract_principle_variation(GAMESTATE *game)
{
   int c;
   int lower, upper, h;
   MOVE move;
   
   game->pv.movecount = 0;
   
   while (retrieve_hashtable(game->hash1, game->hash2, &lower, &upper, &h, &move)) {
      if (h<0 || moves_are_equal(move, null_move))
         break;
      game->pv.move[game->pv.movecount++] = move;
      playmove(game, move);
   }
   
   for (c=0; c<game->pv.movecount; c++)
      takeback(game);
}
