#include "evaluate.h"
#include "board.h"
#include "game.h"
#include "sbr.h"

int evaluate_count;

static int knight_table[64] = {
   0, 2,  4,  6,  6,  4, 2, 0,
   2, 4,  6,  8,  8,  6, 4, 2,
   4, 6,  8, 10, 10,  8, 6, 4,
   6, 8, 10, 12, 12, 10, 8, 6,
   6, 8, 10, 12, 12, 10, 8, 6,
   4, 6,  8, 10, 10,  8, 6, 4,
   2, 4,  6,  8,  8,  6, 4, 2,
   0, 2,  4,  6,  6,  4, 2, 0
};

static int bishop_table[64] = {
   0, 2,  4,  6,  6,  4, 2, 0,
   2, 6,  6,  8,  8,  6, 6, 2,
   4, 6,  8,  8,  8,  8, 6, 4,
   6, 8, 10, 12, 12, 10, 8, 6,
   6, 8, 10, 12, 12, 10, 8, 6,
   4, 6,  8,  8,  8,  8, 6, 4,
   2, 6,  6,  8,  8,  6, 6, 2,
   0, 2,  4,  6,  6,  4, 2, 0
};

static int king_end_table[64] = {
   0, 2,  4,  6,  6,  4, 2, 0,
   2, 4,  6,  8,  8,  6, 4, 2,
   4, 6,  8, 10, 10,  8, 6, 4,
   6, 8, 10, 12, 12, 10, 8, 6,
   6, 8, 10, 12, 12, 10, 8, 6,
   4, 6,  8, 10, 10,  8, 6, 4,
   2, 4,  6,  8,  8,  6, 4, 2,
   0, 2,  4,  6,  6,  4, 2, 0
};


/* MiniMax evaluation function during the opening */
static inline int opening_evaluation(GAMESTATE *game, int player, int opponent)
{
   BITBOARD bb;
   BITBOARD bk;
   int square;
   int ev = 0;

   /* Piece positions */

   /* Centralized pawns */
   bb = and_bitboards(and_bitboards(game->board.bbc[player], game->board.bbp[0]), board_centre);
   if (bb)
      ev+=15;
   bb = and_bitboards(and_bitboards(game->board.bbc[player], game->board.bbp[0]), board_xcentre);
   if (bb)
      ev+=10;

   bb = and_bitboards(and_bitboards(game->board.bbc[opponent], game->board.bbp[0]), board_centre);
   if (bb)
      ev-=15;
   bb = and_bitboards(and_bitboards(game->board.bbc[opponent], game->board.bbp[0]), board_xcentre);
   if (bb)
      ev-=10;

   /* Knights */
   bb = and_bitboards(game->board.bbc[player], game->board.bbp[1]);
   while(bb) {
      square = sbr_longlong(bb);
      ev += knight_table[square];
      unset_bitboard(&bb, square);
   }   
   bb = and_bitboards(game->board.bbc[opponent], game->board.bbp[1]);
   while(bb) {
      square = sbr_longlong(bb);
      ev -= knight_table[square];
      unset_bitboard(&bb, square);
   }

   /* Bishops */
   bb = and_bitboards(game->board.bbc[player], game->board.bbp[2]);
   while(bb) {
      square = sbr_longlong(bb);
      ev += bishop_table[square];
      unset_bitboard(&bb, square);
   }   
   bb = and_bitboards(game->board.bbc[opponent], game->board.bbp[2]);
   while(bb) {
      square = sbr_longlong(bb);
      ev -= bishop_table[square];
      unset_bitboard(&bb, square);
   }
   
   /* Discourage queen sorties */
   bb = and_bitboards(game->board.bbp[4], or_bitboards(board_rank4, board_rank5));
   if (and_bitboards(game->board.bbc[player], bb))
      ev-=12;
   if (and_bitboards(game->board.bbc[opponent], bb))
      ev+=12;

   bk = and_bitboards(game->board.bbp[5], game->board.bbp[6]);
   /* Penalize rooks that have moved while the King has not moved */
   bb = invert_bitboard(and_bitboards(game->board.bbp[3], game->board.bbp[6]));
   if (and_bitboards(bb, game->board.bbp[player]) && 
       and_bitboards(game->board.bbp[player], bk)) {
      ev-=15;
   }

   bb = invert_bitboard(and_bitboards(game->board.bbp[3], game->board.bbp[6]));
   if (and_bitboards(bb, game->board.bbp[opponent]) && 
       and_bitboards(game->board.bbp[opponent], bk)) {
      ev+=15;
   }
   
   /* Penalize if opponent king has castled and own king has not */
   if (and_bitboards(game->board.bbp[player], bk)) {
      ev-=10;
   }
   if (and_bitboards(game->board.bbp[opponent], bk)) {
      ev+=10;
   }
   
   return ev;
}

/* MiniMax evaluation function during the middle game */
static inline int middle_evaluation(GAMESTATE *game, int player, int opponent)
{
   BITBOARD bb;
   int square;
   int ev = 0;

   /* Evaluate bishop pair */
   /* All we do is count the number of bishops, not wether they are actually */
   /*  on different coulours. Such positions are rare anyway though. */
   if (!is_power_of_two(and_bitboards(game->board.bbc[player], game->board.bbp[2])))
      ev+=30;
   if (!is_power_of_two(and_bitboards(game->board.bbc[opponent], game->board.bbp[2])))
      ev-=30;
      
   /* Piece positions */
   /* Knights */
   bb = and_bitboards(game->board.bbc[player], game->board.bbp[1]);
   while(bb) {
      square = sbr_longlong(bb);
      ev += knight_table[square];
      unset_bitboard(&bb, square);
   }   
   bb = and_bitboards(game->board.bbc[opponent], game->board.bbp[1]);
   while(bb) {
      square = sbr_longlong(bb);
      ev -= knight_table[square];
      unset_bitboard(&bb, square);
   }

   /* Bishops */
   bb = and_bitboards(game->board.bbc[player], game->board.bbp[2]);
   while(bb) {
      square = sbr_longlong(bb);
      ev += bishop_table[square];
      unset_bitboard(&bb, square);
   }   
   bb = and_bitboards(game->board.bbc[opponent], game->board.bbp[2]);
   while(bb) {
      square = sbr_longlong(bb);
      ev -= bishop_table[square];
      unset_bitboard(&bb, square);
   }
   
   /* Centralized pawns */
   bb = and_bitboards(and_bitboards(game->board.bbc[player], game->board.bbp[0]), board_centre);
   if (bb)
      ev+=15;
   bb = and_bitboards(and_bitboards(game->board.bbc[player], game->board.bbp[0]), board_xcentre);
   if (bb)
      ev+=10;

   bb = and_bitboards(and_bitboards(game->board.bbc[opponent], game->board.bbp[0]), board_centre);
   if (bb)
      ev-=15;
   bb = and_bitboards(and_bitboards(game->board.bbc[opponent], game->board.bbp[0]), board_xcentre);
   if (bb)
      ev-=10;
      
   return ev;
}

/* MiniMax evaluation function during the end game */
static inline int ending_evaluation(GAMESTATE *game, int player, int opponent)
{
   BITBOARD bb;
   BITBOARD bp;
   int square;
   int ev = 0;
   int n;
   /* In the middle game, the King should be at the centre of the board */

   bb = and_bitboards(game->board.bbc[player], game->board.bbp[5]);
   while(bb) {
      square = sbr_longlong(bb);
      ev += king_end_table[square];
      unset_bitboard(&bb, square);
   }   
   bb = and_bitboards(game->board.bbc[opponent], game->board.bbp[5]);
   while(bb) {
      square = sbr_longlong(bb);
      ev -= king_end_table[square];
      unset_bitboard(&bb, square);
   }
   
   /* Detect free pawns */
   
   /* White */
   bb = and_bitboards(game->board.bbp[0], game->board.bbc[1]);
   for (n=0; n<6; n++)
      bb = or_bitboards(bb, shift_bitboard_row_down(bb));
   bb = invert_bitboard(bb);
   bp = and_bitboards(game->board.bbp[0], game->board.bbc[0]);
   if (bp && and_bitboards(bb, bp)) {
      if (player==0) {
         ev += 20*(sbr_longlong(bp)>>3);
      } else {
         ev -= 20*(sbr_longlong(bp)>>3);
      }
   }
   
   /* Black */
   bb = and_bitboards(game->board.bbp[0], game->board.bbc[0]);
   for (n=0; n<6; n++)
      bb = or_bitboards(bb, shift_bitboard_row_up(bb));
   bb = invert_bitboard(bb);
   bp = and_bitboards(game->board.bbp[0], game->board.bbc[1]);
   if (bp && and_bitboards(bb, bp)) {
      if (player==1) {
         ev += 20*(sbr_longlong(bp)>>3);
      } else {
         ev -= 20*(sbr_longlong(bp)>>3);
      }
   }
   
   return ev;
}

/* MiniMax evaluation function */
/* Returns >0 if the position favours the player to move, */
/*         <0 if it favours the opponent */
int evaluate(GAMESTATE *game)
{
   int ev = 0;
   int player;
   int opponent;
   
   evaluate_count++;
   
   /* If the opponent of the player to move is being checked, then claim */
   /*  that we have found mate */
   /*
   if (check(&game->board, game->player^BLACK)) {
      return piece_values[5];
   }
   */
   
   player = game->player >> 7;
   opponent = player^0x01;
   
   /* Assess material balance */
   ev = game->material[player]-game->material[opponent];
   
   /* Fine tune */
   if (and_bitboards(and_bitboards(game->board.bbc[player], game->board.bbp[5]), game->board.bbp[6])
       ||
       and_bitboards(and_bitboards(game->board.bbc[opponent], game->board.bbp[5]), game->board.bbp[6])) {

      /* The opening: this is before either king has castled */
      ev += opening_evaluation(game, player, opponent);
   } else if (game->material[player]+game->material[opponent] < ENDGAME_MATERIAL) {
      /* The end game, this is when the total material on the board drops */
      /*  below some fixed number */
      ev += ending_evaluation(game, player, opponent);
   } else {
      /* If it's neither opening nor end game, it's middle game */
      ev += middle_evaluation(game, player, opponent);
   }
   
   return ev;
}
