//#define DEBUGMODE
#include <string.h>
#include "hashtab.h"
#include "assert.h"
#include "move.h"

#ifdef USE_PTHREADS
   #include <pthread.h>
#endif

#define RETRIEVE_DISTANCE     20

typedef struct TABLE_ENTRY {
   unsigned int alt_key;   /* Second hash key */
   unsigned int lower;     /* Lower bound on minimax value */
   unsigned int upper;     /* Upper bound on minimax value */
   unsigned int horizon;   /* Horizon distance */
   MOVE move;              /* Best move at this node */
} TABLE_ENTRY;

static TABLE_ENTRY *table = NULL;
static unsigned int table_size = 0;

int hash_table_hits;
int hash_table_misses;
int hash_table_collissions;

/* In multi-threaded programs, we need to make sure only one thread */
/*  at a time changes the hashtable */
#ifdef USE_PTHREADS
static pthread_mutex_t hashtable_lock = PTHREAD_MUTEX_INITIALIZER;
#define LOCK()    pthread_mutex_lock(&hashtable_lock)
#define UNLOCK()  pthread_mutex_unlock(&hashtable_lock)
#else
#define LOCK()
#define UNLOCK()
#endif

/* Initialize a hashtable with size elements. Returns size on succes, 0 */
/*  on failure. */
int initialize_hashtable(int size)
{
   LOCK();
   
   table = malloc(size * sizeof *table);
   if (table) {
      table_size = size;
      UNLOCK();
      empty_hashtable();
      LOCK();
      trace("Hash table allocated");
   }
   
   hash_table_hits = 0;
   hash_table_misses = 0;
   hash_table_collissions = 0;
   
   UNLOCK();
   
   return table_size;
}

void destroy_hashtable(void)
{
   LOCK();
   
   free(table);
   table = NULL;
   table_size = 0;
   
   UNLOCK();
}

void empty_hashtable(void)
{
   LOCK();
   
   memset(table, 0, table_size*sizeof *table);
   
   UNLOCK();
}

void inc_hashtable_horizon(void)
{
}

void dec_hashtable_horizon(void)
{
   int n;
      
   for(n=0; n<table_size; n++) {
      LOCK();
      table[n].horizon--;
      UNLOCK();
   }
}

/* Returns non-zero if the position was stored in the table */
int store_hashtable(unsigned int key1, unsigned int key2, int lower, int upper, int horizon, MOVE move)
{
   int n;
   int i;
   
   if (!table_size)
      return 0;
   
   LOCK();
   
   assert(key1 < table_size);
   assert(key2 < table_size);
   
   key1 &= table_size-1;
   key2 &= table_size-1;
   
   /* See if position is free, or entry is already there */
   if (table[key1].alt_key == 0 || table[key1].alt_key==key2 || table[key1].horizon<-4) {
      table[key1].alt_key = key2;
      table[key1].lower = lower;
      table[key1].upper = upper;
      table[key1].horizon = horizon;
      table[key1].move = move;
      
      UNLOCK();
      
      return 1;
   }
   
   /* Try to store at a different location, at most RETRIEVE_DISTANCE places */
   /*  above the first hash key. */
   for (n=0; n<RETRIEVE_DISTANCE; n++) {
      i = (key1+n) & (table_size-1);
      if (table[i].alt_key == 0 || table[i].alt_key==key2 || table[i].horizon<-4) {
         table[i].alt_key = key2;
         table[i].lower = lower;
         table[i].upper = upper;
         table[i].horizon = horizon;
         table[i].move = move;
         
         UNLOCK();
      
         return 1;
      }
   }

   /* No luck - override the alien at the first hash lock */
   table[key1].alt_key = key2;
   table[key1].lower = lower;
   table[key1].upper = upper;
   table[key1].horizon = horizon;
   table[key1].move = move;
      
   hash_table_collissions++;
   
   UNLOCK();
   
   return 1;
}

/* Returns non-zero if the data could be found in the table */
int retrieve_hashtable(unsigned int key1, unsigned int key2, int *lower, int *upper, int *horizon, MOVE *move)
{
   int n;
   int i;
   if (!table_size)
      return 0;

   LOCK();

   assert(key1 < table_size);
   assert(key2 < table_size);
   
   key1 &= table_size-1;
   key2 &= table_size-1;
   
   if (table[key1].alt_key==key2) {
      /* Found at first entry */
      *lower = table[key1].lower;
      *upper = table[key1].upper;
      *horizon = table[key1].horizon;
      *move = table[key1].move;
      trace("Table hit");
      hash_table_hits++;
      
      UNLOCK();
      return 1;
   }

   /* See if stored at a different location, at most RETRIEVE_DISTANCE places */
   /*  above the first hash key. */
   for (n=0; n<RETRIEVE_DISTANCE; n++) {
      i = (key1+n) & (table_size-1);
      if (table[i].alt_key==key2) {
         /* Found */
         *lower = table[i].lower;
         *upper = table[i].upper;
         *horizon = table[i].horizon;
         *move = table[i].move;
         trace("Table hit");
         hash_table_hits++;
         UNLOCK();
         return 1;
      }
   }
   
   hash_table_misses++;
   
   trace("Retrieve failed");
   UNLOCK();
   return 0;
}

void write_hashtable(char *filename)
{
   FILE *f;
   
   LOCK();
   f=fopen(filename, "w");
   fwrite(table, sizeof *table, table_size, f);
   fclose(f);
   UNLOCK();
}

void load_hashtable(char *filename)
{
   FILE *f;
   
   LOCK();
   f=fopen(filename, "r");
   if (f)
      fread(table, sizeof *table, table_size, f);
   fclose(f);
   UNLOCK();
}
