#include "square.h"
#include "board.h"
#include "move.h"
#include "game.h"

#define PIECESTRING   " NBRQK"

/* Wrappers to be replaced by custom memory management code in due time */
static void *alloc_position_data(void)
{
   return malloc(sizeof(GAME_POSITION_DATA));
}

static void free_position_data(void *data)
{
   free(data);
}

void playgame_init(void)
{
   init_lists();
}

void playgame_shotdown(void)
{
   free_lists();
}

/* Initialize the game datastructure to correspond to a new game */
void start_newgame(GAMESTATE *game)
{
   /* Load the starting position */
   clear_board(&(game->board));
   begin_board(&(game->board));

   game->player = WHITE;
   count_material(game);
   
   game->movelist = NULL;
   game->moves_played = 0;

   game->hash1 = calculate_key1(&game->board, game->player);
   game->hash2 = calculate_key2(&game->board, game->player);
}

/* Frees the memory used by a gamestate */
void end_game(GAMESTATE *game)
{
   if (game->movelist) {
      /* Move to the start of the list */
      while(game->movelist->prev)
         game->movelist = game->movelist->prev;
   }
   destroy_list(game->movelist, free_position_data);
   game->movelist = NULL;
   game->movelist_tail = NULL;
   game->moves_played = 0;
}

/* Counts the material on the board */
/* Only needed to initialize, since the movement code will keep track of it */
void count_material(GAMESTATE *game)
{
   int c;
   int p;
   
   game->material[0] = 0;
   game->material[1] = 0;
   
   for(c=0; c<64; c++) {
      p = get_piece(&game->board, c);
      if (p&TIMBER) {
         game->material[(p&BLACK)>>7] += piece_values[sbr_longlong(p&TIMBER)];
      }
   }
}

/* Performs a move on the board and updates the position, but not the */
/*  en-passant status */
void forward_move(BOARD *board, const MOVE move)
{
   /* Check special moves */
   if (!is_castling(move)) {
      /* Remove the piece from the original square */
      remove_piece(board, get_origin(move));

      if (!is_promotion(move)) {
         if (!is_enpassant(move)) {     /* Ordinary move */
            /* Place the piece on the destination square, disabeling the ini-bit */
            remove_piece(board, get_destination(move));
            place_piece(board, get_destination(move), get_move_piece(move)&(0xFF^INIBIT));
         } else {                      /* En-passant */
            /* Place the pawn on the destination square */
            place_piece(board, get_destination(move), get_move_piece(move)&(PAWN|BLACK));
            /* Remove the captured pawn */
            remove_piece(board, get_ep_square(move));
         }
      } else {                         /* Promotion */
         /* Place the new piece on the destination square */
         remove_piece(board, get_destination(move));
         place_piece(board, get_destination(move), get_move_piece(move)&(QUEEN|ROOK|BISHOP|KNIGHT|BLACK));
      }
   } else {                            /* Castling */
      /* Remove the king */
      remove_piece(board, get_origin(move));
      /* Remove the rook */
      remove_piece(board, get_rook_origin(move));
      /* Move the king */
      place_piece(board, get_destination(move), get_move_piece(move)&(KING|BLACK));
      /* Move the rook */
      place_piece(board, get_rook_destination(move), get_move_piece(move)&(ROOK|BLACK));
   }
}

/* Inverts the move on the board, but does not restore the EP state */
void backward_move(BOARD *board, const MOVE move)
{
   /* Check special moves */
   if (!is_castling(move)) {
      if (!is_promotion(move)) {
         if (!is_enpassant(move)) {    /* Ordinary move */
            remove_piece(board, get_destination(move));
            /* Return captured piece to its original position */
            if (get_captured_piece(move)) {
               place_piece(board, get_destination(move), get_captured_piece(move));
            }
            /* Return moved piece to its original position */
            place_piece(board, get_origin(move), get_move_piece(move));
         } else {                      /* En-passant */
            /* Save the en-passant square in the board structure */
            board->ep_square = get_ep_square(move);
            /* Return captured pawn to its original position */
            place_piece(board, get_ep_square(move), get_captured_piece(move));
            /* Return capturing pawn to its original position */
            place_piece(board, get_origin(move), get_move_piece(move)&(PAWN|BLACK));
            remove_piece(board, get_destination(move));
         }
      } else {                         /* Promotion */
         remove_piece(board, get_destination(move));
         /* Return captured piece to its original position */
         if (get_captured_piece(move)) {
            place_piece(board, get_destination(move), get_captured_piece(move));
         }
         /* Return pawn to its original position */
         place_piece(board, get_origin(move), get_move_piece(move)&(PAWN|BLACK));
      }
   } else {                            /* Castling */
      /* Remove the king and the rook */
      remove_piece(board, get_destination(move));
      remove_piece(board, get_rook_destination(move));
      /* Put back in original position, restoring castling rights */
      place_piece(board, get_origin(move), (get_move_piece(move)&(KING|BLACK))|INIBIT);
      place_piece(board, get_rook_origin(move), (get_move_piece(move)&(ROOK|BLACK))|INIBIT);
   }
}

/* Returns the depth-last move by player in game */
MOVE get_move_history(GAMESTATE *game, int player, int depth)
{
   GAME_POSITION_DATA *gpd;
   int move_back;
   LINK *l;
   
   
   /* Compute how many move we need to traverse back into the tree */
   move_back = 2*depth + (game->player==player);
   
   /* That many played? */
   if (game->moves_played>move_back) {
      l = game->movelist_tail;
      while (move_back) {
         l = l->prev;
         move_back--;
      }
      gpd = l->data;
      return gpd->move;
   } else {
      return null_move;
   }
}

void playmove(GAMESTATE *game, MOVE move)
{
   GAME_POSITION_DATA *gpd;
   
   gpd = alloc_position_data();
   
   if (!gpd) {
      fprintf(stderr, "Out of memory during play - can't continue!\n");
      abort();
   }

   /* Store move and en-passant information */   
   gpd->move = move;
   gpd->ep_square = game->board.ep_square;

   game->movelist = append_to_list(gpd, game->movelist);
   game->moves_played++;
   
   if (game->movelist_tail) {
      game->movelist_tail = game->movelist_tail->next;
   } else {
      game->movelist_tail = game->movelist;
   }

   /* Now update the actual board position */
   forward_move(&(game->board), move);
   /* Set en-passant field (needed for move generation) */
   if (is_doublepawn(move)) {
      game->board.ep_square = get_destination(move);
   } else {
      game->board.ep_square = 0;
   }
   game->player ^= BLACK;
   /* Now adjust material, if needed */
   if (is_capture(move)) {
      int p = get_captured_piece(move);
      game->material[(p&BLACK)>>7] -= piece_values[sbr_longlong(p&TIMBER)];
   }
   if (is_promotion(move)) {
      int p = get_move_piece(move);
      game->material[(p&BLACK)>>7] += piece_values[sbr_longlong(p&TIMBER)]-piece_values[0];
   }

   /* Update hashkeys */
   game->hash1 = calculate_key1(&game->board, game->player);
   game->hash2 = calculate_key2(&game->board, game->player);
}

/* Reverts the last move played */
void takeback(GAMESTATE *game)
{
   GAME_POSITION_DATA *gpd;
   LINK *l;
   
   if (game->moves_played) {
      l = game->movelist_tail;
      gpd = l->data;
      backward_move(&(game->board), gpd->move);

      /* Now adjust material, if needed */
      if (is_capture(gpd->move)) {
         int p = get_captured_piece(gpd->move);
         game->material[(p&BLACK)>>7] += piece_values[sbr_longlong(p&TIMBER)];
      }
      if (is_promotion(gpd->move)) {
         int p = get_move_piece(gpd->move);
         game->material[(p&BLACK)>>7] -= piece_values[sbr_longlong(p&TIMBER)]-piece_values[0];
      }

      game->board.ep_square = gpd->ep_square;
      game->movelist_tail = game->movelist_tail->prev;
      if (l->prev) {
         l->prev->next = NULL;
      } else {
         game->movelist = NULL;
      }
      free_position_data(gpd);
      free_link(l);
      
      game->moves_played--;

      game->player ^= BLACK;
   }

   /* Update hashkeys */
   game->hash1 = calculate_key1(&game->board, game->player);
   game->hash2 = calculate_key2(&game->board, game->player);
}

char *get_move_string_algebraic(const BOARD *b, const MOVE m, char *s)
{
   if (s) {
      int piece = get_move_piece(m) & PIECE;  
      int square = get_origin(m);
      int dest = get_destination(m);
      int srow = unpack_row(square);
      int scolumn = unpack_column(square);
      int drow = unpack_row(dest);
      int dcolumn = unpack_column(dest);
      sprintf(s, "%c%c%d%c%c%d",
             PIECESTRING[sbr(piece)],
             scolumn+'a', srow+1,
             is_capture(m)?'x':'-',
             dcolumn+'a', drow+1);
   }
   return s;
}
