#include <string.h>
#include "movegen.h"
#include "bitboard.h"
#include "board.h"
#include "pieces.h"
#include "itype.h"
#include "move.h"
#include "sbr.h"

/* For the King and Knight, movement tables are stored in memory in bitboards */
static BITBOARD king_moves[64];
static BITBOARD knight_moves[64];

void calculate_move_tables(void)
{
   int c;
   int x, y, dx, dy;
   
   for (c=0; c<64; c++) {
      x = c&7;
      y = c>>3;

      king_moves[c] = board_empty;
      for (dx=-1; dx<2; dx++)
         for (dy=-1; dy<2; dy++) {
            if ( (dx||dy) && (x+dx>=0 && x+dx<=7 && y+dy>=0 && y+dy<=7) ) {
               set_bitboard(&(king_moves[c]), (x+dx) + ((y+dy)<<3));
            }
         }

      knight_moves[c] = board_empty;
      for (dx=-3; dx<4; dx++)
         for (dy=-3; dy<4; dy++) {
            if ( (dx*dx + dy*dy == 5)  && (x+dx>=0 && x+dx<=7 && y+dy>=0 && y+dy<=7) ) {
               set_bitboard(&(knight_moves[c]), (x+dx) + ((y+dy)<<3));
            }
         }
   }
}

void print_move_list(MOVE_GEN_STRUCT *ml, int *nummoves, char ***movestring)
{
   int c;
   int piece;
   int square;
   int dest;
   
   assert(nummoves);
   assert(movestring);
   
   (*nummoves) = ml->total.nummoves;
   (*movestring) = malloc(ml->total.nummoves * sizeof *movestring);
   
   for (c=0; c<ml->total.nummoves; c++) {
      (*movestring)[c] = malloc(10);
      piece = get_move_piece(ml->total.move[c]) & PIECE;  
      square = get_origin(ml->total.move[c]);
      dest = get_destination(ml->total.move[c]);
      if (!is_castling(ml->total.move[c]) && !is_promotion(ml->total.move[c])) {
         if (is_enpassant(ml->total.move[c])) {
            sprintf((*movestring)[c], "%c%cx%c%cep",
                   (square&7)+'a', (square>>3)+'1',
                   (dest&7)+'a', (dest>>3)+'1');
         } else {
            sprintf((*movestring)[c], "%c%c%c%c%c%c ",
                   "PNBRQK"[sbr(piece)],
                   (square&7)+'a', (square>>3)+'1',
                   is_capture(ml->total.move[c])?'x':'-',
                   (dest&7)+'a', (dest>>3)+'1');
         }
      } else if (is_castling(ml->total.move[c])) {
         if (dest>square)
            sprintf((*movestring)[c], "   O-O");
         else
            sprintf((*movestring)[c], " O-O-O");
      } else if (is_promotion(ml->total.move[c])) {
         sprintf((*movestring)[c], "%c%c%c%c%c%c ",
                (square&7)+'a', (square>>3)+'1',
                is_capture(ml->total.move[c])?'x':'-',
                (dest&7)+'a', (dest>>3)+'1',
                "PNBRQK"[sbr(piece)]);
      }
   }
}

/* Helper to generate moves for fast-paced pieces (queen, rook, bishop): */
/*  the code is practically the same in all cases */
#define generate_fast_move(stepgen) \
{                                                              \
   BITBOARD bb = stepgen(bp);                                  \
   uint8 ptaken;                                               \
   int dest;                                                   \
   /* Generate normal moves until we hit a non-empty square */ \
   /*  or the edge of the board is reached */                  \
   while(and_bitboards(bb, empty)) {                           \
      covered = or_bitboards(covered, bb);                     \
                                                               \
      if (and_bitboards(bb, tmask)) {                          \
         encode_move(ml->normal.move+ml->normal.nummoves,      \
                     piece, square, sbr_longlong(bb), NONE);  \
         ml->normal.nummoves++;                                \
      }                                                        \
                                                               \
      bb = stepgen(bb);                                        \
   }                                                           \
   covered = or_bitboards(covered, bb);                        \
                                                               \
   /* Check if we have hit a piece we can capture */           \
   if (and_bitboards(and_bitboards(bb, tmask), enemy)) {       \
      dest = sbr_longlong(bb);                                \
      ptaken = get_piece(board, dest);                         \
      encode_move(ml->capture.move+ml->capture.nummoves,       \
                  piece, square, dest, ptaken);                \
      ml->capture.nummoves++;                                  \
   }                                                           \
}

/* Helper to generate pawn moves, except double moves and en-passant moves */
#define generate_pawn_single_move(movetype,disp)               \
{                                                              \
      uint8 ptaken;                                            \
      int dest;                                                \
                                                               \
      /* Promotions */                                         \
      while(promote_pawns) {                                   \
         dest = sbr_longlong(promote_pawns);                  \
         piece = get_piece(board, dest + disp);                \
         ptaken = get_piece(board, dest);                      \
                                                               \
         if (and_bitboards(promote_pawns, tmask)) {            \
            encode_move(ml->promotion.move+ml->promotion.nummoves,\
                        piece|QUEEN, dest + disp, dest, ptaken);  \
            ml->promotion.nummoves++;                             \
            encode_move(ml->promotion.move+ml->promotion.nummoves,\
                        piece|ROOK, dest + disp, dest, ptaken);   \
            ml->promotion.nummoves++;                             \
            encode_move(ml->promotion.move+ml->promotion.nummoves,\
                        piece|BISHOP, dest + disp, dest, ptaken); \
            ml->promotion.nummoves++;                             \
            encode_move(ml->promotion.move+ml->promotion.nummoves,\
                        piece|KNIGHT, dest + disp, dest, ptaken); \
            ml->promotion.nummoves++;                             \
         }                                                     \
                                                               \
         bp = board_empty;                                     \
         set_bitboard(&bp, dest);                              \
         promote_pawns = xor_bitboards(promote_pawns, bp);     \
      }                                                        \
      /* Normal moves */                                       \
      while(bb) {                                              \
         dest = sbr_longlong(bb);                             \
         piece = get_piece(board, dest + disp);                \
         ptaken = get_piece(board, dest);                      \
                                                               \
         if (and_bitboards(bb, tmask)) {                       \
            if (dest == ep_dest) {       /* En-passant */      \
               piece |= ENPASSANT_MOVE;                        \
               ptaken = get_piece(board, board->ep_square);    \
            }                                                  \
                                                               \
            encode_move(ml->movetype.move+ml->movetype.nummoves,\
                        piece, dest + disp, dest, ptaken);     \
            ml->movetype.nummoves++;                       \
         }                                                     \
                                                               \
         bp = board_empty;                                     \
         set_bitboard(&bp, dest);                              \
         bb = xor_bitboards(bb, bp);                           \
      }                                                        \
}

/* Helper to generate pawn moves */
#define generate_pawn_moves(shift_f,shift_l,shift_r,sign,dl,dr,rank3,rank8)\
{                                                                       \
   pawns = and_bitboards(own, board->bbp[0]);                           \
   /* Keep track of what squares are covered by the pawns */            \
   covered = or_bitboards(covered, shift_r(pawns));                 \
   covered = or_bitboards(covered, shift_l(pawns));                 \
                                                                        \
   bb = and_bitboards(empty, shift_f(pawns));                       \
   /* Keep seperate track of pawns that can do a double move or promote */\
   /*  Only pawns that can move to the third row could move to the fourth */\
   double_pawns = and_bitboards(empty, shift_f(and_bitboards(bb, rank3)));\
   promote_pawns = and_bitboards(rank8, bb);                        \
   /* Do not double check pawns we already check for promotion */       \
   bb = xor_bitboards(bb, promote_pawns);                               \
   /* Now generate pawn moves */                                        \
   /* Double steps */                                                   \
   while(double_pawns) {                                                \
      int dest;                                                         \
                                                                        \
      dest = sbr_longlong(double_pawns);                               \
      if (and_bitboards(bb, tmask)) {                                   \
         piece = get_piece(board, dest sign 16);                    \
                                                                        \
         encode_move(ml->normal.move+ml->normal.nummoves,               \
                     piece, dest sign 16, dest, NONE);              \
         ml->normal.nummoves++;                                         \
      }                                                                 \
                                                                        \
      bp = board_empty;                                                 \
      set_bitboard(&bp, dest);                                          \
      double_pawns = xor_bitboards(double_pawns, bp);                   \
   }                                                                    \
                                                                        \
   generate_pawn_single_move(normal, sign 8);                       \
   /* Calculate capture moves to the right, including en-passant */     \
   bb = and_bitboards(or_bitboards(enemy, en_passant), shift_r(pawns));\
   promote_pawns = and_bitboards(rank8, bb);                        \
   bb = xor_bitboards(bb, promote_pawns);                               \
   generate_pawn_single_move(capture, sign dr);                     \
   /* Calculate capture moves to the left , including en-passant */     \
   bb = and_bitboards(or_bitboards(enemy, en_passant), shift_l(pawns));\
   promote_pawns = and_bitboards(rank8, bb);                        \
   bb = xor_bitboards(bb, promote_pawns);                               \
   generate_pawn_single_move(capture, sign dl);                     \
}

/* Move generation work horse - can also be tailored to generate only */
/*  certain types of moves by including a mask. */
/* Only pieces that obey the fmask will be considered, */
/*  only destinations in the tmask will be considered */
static void move_gen_mask(BITBOARD fmask, BITBOARD tmask, BOARD *board, int player, MOVE_GEN_STRUCT *ml)
{
   int piece;
   int square;
   int ep_dest;
   BITBOARD own;
   BITBOARD enemy;
   BITBOARD occupied;
   BITBOARD empty;
   BITBOARD pawns;
   BITBOARD double_pawns;
   BITBOARD promote_pawns;
   BITBOARD en_passant;
   BITBOARD bp;
   BITBOARD bb;
   BITBOARD bk;
   BITBOARD covered;

   /* Initilize appropriate bitboards */
   /* The covered bitboard eventually will hold all squares that the current */
   /*  player can attack */
   covered = board_empty;
   en_passant = board_empty;
   ep_dest = 0;
   if (player == WHITE) {
      own = board->bbc[0];
      enemy = board->bbc[1];
      /* Put a fake piece on the en-passant square */
      if (board->ep_square) {
         ep_dest = board->ep_square+8;
         set_bitboard(&en_passant, ep_dest);
      }
   } else {
      own = board->bbc[1];
      enemy = board->bbc[0];
      /* Put a fake piece on the en-passant square */
      if (board->ep_square) {
         ep_dest = board->ep_square-8;
         set_bitboard(&en_passant, ep_dest);
      }
   }

   occupied = or_bitboards(own, enemy);
   empty = invert_bitboard(occupied);

   /* Now apply bitmasks */
   //empty = and_bitboards(empty, tmask);
   //occupied = invert_bitboard(empty);

   own = and_bitboards(own, fmask);
   enemy = and_bitboards(enemy, tmask);
   
   /* Return if we can't generate any moves with these masks */
   if (!own || (!empty && !enemy))
      return;
   
   /* Generate rook moves and straight queen moves */
   bb = and_bitboards(own, or_bitboards(board->bbp[3], board->bbp[4]));
   while(bb) {
      square = sbr_longlong(bb);
      piece = get_piece(board, square);
      
      bp = board_empty;
      set_bitboard(&bp, square);
      
      generate_fast_move(shift_bitboard_row_up);
      generate_fast_move(shift_bitboard_row_down);
      generate_fast_move(shift_bitboard_file_right);
      generate_fast_move(shift_bitboard_file_left);
      
      bb = xor_bitboards(bb, bp);
   }

   /* Generate bishop moves and diagonal queen moves */
   bb = and_bitboards(own, or_bitboards(board->bbp[2], board->bbp[4]));
   while(bb) {
      square = sbr_longlong(bb);
      piece = get_piece(board, square);
      
      bp = board_empty;
      set_bitboard(&bp, square);
      
      generate_fast_move(shift_bitboard_right_up);
      generate_fast_move(shift_bitboard_right_down);
      generate_fast_move(shift_bitboard_left_up);
      generate_fast_move(shift_bitboard_left_down);
      
      bb = xor_bitboards(bb, bp);
   }

   /* Generate Knight moves */
   bk = and_bitboards(own, board->bbp[1]);
   
   while(bk) {
      square = sbr_longlong(bk);
      piece = get_piece(board, square);
      covered = or_bitboards(covered, knight_moves[square]);
      /* Captures */
      bb = and_bitboards(enemy, knight_moves[square]);
      while(bb) {
         uint8 ptaken;
         int dest;
      
         dest = sbr_longlong(bb);
         ptaken = get_piece(board, dest);
         
         if (and_bitboards(bb, tmask)) {
            encode_move(ml->capture.move+ml->capture.nummoves,
                        piece, square, dest, ptaken);
            ml->capture.nummoves++;
         }
         
         bp = board_empty;
         set_bitboard(&bp, dest);
      
         bb = xor_bitboards(bb, bp);
      }
      /* Normal moves */
      bb = and_bitboards(empty, knight_moves[square]);
      while(bb) {
         int dest;
         dest = sbr_longlong(bb);
      
         if (and_bitboards(bb, tmask)) {
            encode_move(ml->normal.move+ml->normal.nummoves,
                        piece, square, dest, NONE);
            ml->normal.nummoves++;
         }

         bp = board_empty;
         set_bitboard(&bp, dest);
      
         bb = xor_bitboards(bb, bp);
      }
      bp = board_empty;
      set_bitboard(&bp, square);
      bk = xor_bitboards(bk, bp);
   }

   /* Generate King moves */
   bb = and_bitboards(own, board->bbp[5]);
   square = sbr_longlong(bb);
   piece = get_piece(board, square);
   covered = or_bitboards(covered, king_moves[square]);
   /* Captures */
   bb = and_bitboards(enemy, king_moves[square]);
   while(bb) {
      uint8 ptaken;
      int dest;
      
      dest = sbr_longlong(bb);
      ptaken = get_piece(board, dest);

      if (and_bitboards(bb, tmask)) {
         encode_move(ml->capture.move+ml->capture.nummoves,
                     piece, square, dest, ptaken);
         ml->capture.nummoves++;
      }
      
      bp = board_empty;
      set_bitboard(&bp, dest);
      
      bb = xor_bitboards(bb, bp);
   }
   /* Normal moves */
   bb = and_bitboards(empty, king_moves[square]);
   while(bb) {
      int dest;
      dest = sbr_longlong(bb);
      
      if (and_bitboards(bb, tmask)) {
         encode_move(ml->normal.move+ml->normal.nummoves,
                     piece, square, dest, NONE);
         ml->normal.nummoves++;
      }
      
      bp = board_empty;
      set_bitboard(&bp, dest);
      
      bb = xor_bitboards(bb, bp);
   }
   /* Castling */
   bp = and_bitboards(own, board->bbp[5]);
   bb = and_bitboards(own, and_bitboards(board->bbp[6], or_bitboards(board->bbp[5], board->bbp[3])));
   /* Kingside */
   if (bb & bb<<3 && (empty&(bp<<1)) && (empty&(bp<<2))) {
      /* Make sure we're not in check and need to pass the rook square */
      if (!check(board, player) && !covers(board, player^BLACK, square+1)) {
         encode_move(ml->normal.move+ml->normal.nummoves,
                     piece|CASTLE_MOVE, square, square+2, NONE);
      ml->normal.nummoves++;
      }
   }
   /* Queenside */
   if (bb & bb>>4 && (empty&(bp>>1)) && (empty&(bp>>2)) && (empty&(bp>>3))) {
      if (!check(board, player) && !covers(board, player^BLACK, square-1)) {
         encode_move(ml->normal.move+ml->normal.nummoves,
                     piece|CASTLE_MOVE, square, square-2, NONE);
         ml->normal.nummoves++;
      }
   }
   
   /* Generate pawn moves */
   if (player==WHITE) {
      generate_pawn_moves(shift_bitboard_row_up,
                          shift_bitboard_right_up,
                          shift_bitboard_left_up, 
                          -,9,7,board_rank3,board_rank8);
   } else {
      generate_pawn_moves(shift_bitboard_row_down,
                          shift_bitboard_right_down,
                          shift_bitboard_left_down,
                          +, 7, 9, board_rank6, board_rank1);
   }
}

void move_gen_start(BOARD *board, int player, MOVE_GEN_STRUCT *ml)
{
   assert(ml);
   
   /* Make sure the move list is empty */
   ml->total.nummoves = 0;
   ml->total.cmove = 0;

   ml->capture.nummoves = 0;
   ml->capture.cmove = 0;

   ml->promotion.nummoves = 0;
   ml->promotion.cmove = 0;
   
   ml->normal.nummoves = 0;
   ml->normal.cmove = 0;
   
   /* The following rule only allows rules towards the centre of the board */
   //move_gen_mask(board_all, board_centre|board_xcentre, board, player, ml);

   /* The following rule only allows rules towards the king-side the board */
   //move_gen_mask(board_all, board_kwing, board, player, ml);

   /* The following rule only allows rules from the queen to the king-side the board */
   //move_gen_mask(board_qwing, board_kwing, board, player, ml);

   /* Only allow capture moves */
   move_gen_mask(board_all, board->bbc[1-(player>>7)], board, player, ml);

   /* Disallow capture moves */
   move_gen_mask(board_all, invert_bitboard(board->bbc[1-(player>>7)]), board, player, ml);

   /* Allow all moves */
   //move_gen_mask(board_all, board_all, board, player, ml);


   /* Copy move list */
   if (ml->capture.nummoves) {
      memcpy(ml->total.move+ml->total.nummoves, 
             ml->capture.move, ml->capture.nummoves * sizeof *ml->total.move);
      ml->total.nummoves += ml->capture.nummoves;
   }

   if (ml->promotion.nummoves) {
      memcpy(ml->total.move+ml->total.nummoves, 
             ml->promotion.move, ml->promotion.nummoves * sizeof *ml->total.move);
      ml->total.nummoves += ml->promotion.nummoves;
   }

   if (ml->normal.nummoves) {
      memcpy(ml->total.move+ml->total.nummoves, 
             ml->normal.move, ml->normal.nummoves * sizeof *ml->total.move);
      ml->total.nummoves += ml->normal.nummoves;
   }
}

/* Retrieves the next move from the list. Returns the number of moves that */
/*  were in the list, in other words, returns 0 when there are no more */
/*  valid moves left after this one. */
int get_next_move(MOVE_GEN_STRUCT *ml, MOVE *m)
{
   if (ml->total.cmove<ml->total.nummoves) {
      (*m) = ml->total.move[ml->total.cmove];
      return ml->total.nummoves - (ml->total.cmove++);
   }
   
   return ml->total.nummoves - ml->total.cmove;
}

/* Determines wether or not a move is in the list of valid moves */
int is_valid_move(MOVE m, MOVE_GEN_STRUCT *ml)
{
   int c;
   
   for (c=0; c<ml->total.nummoves; c++) {
      if (moves_are_equal(m, ml->total.move[c]))
         return 1;
   }
   
   return 0;
}

/* Helper function, used to determine wether or not a piece at a distance */
/*  covers this field. The bp bitboard is set to a mask of pieces to check for */
#define scan_direction(stepgen)                                \
{                                                              \
   bb = stepgen(own);                                          \
   /* Generate normal moves until we hit a non-empty square */ \
   /*  or the edge of the board is reached */                  \
   while(and_bitboards(bb, empty)) {                           \
      bb = stepgen(bb);                                        \
   }                                                           \
   if (and_bitboards(bb, bp))                                  \
      return 1;                                                \
}   
/* Returns TRUE if the player attacks the square on the board              */
/* Really a stripped-down version of the move generator.                   */
/*  (yes, of course there are better ways to deal with this, such as       */
/*   keeping track of it in the move generator, something we can at least  */
/*   do for the active player in the current ply).                         */
int covers(BOARD *board, int player, int square)
{
   BITBOARD own;
   BITBOARD enemy;
   BITBOARD occupied;
   BITBOARD empty;
   BITBOARD pawns;
   BITBOARD bp;
   BITBOARD bb;

   /* Initilize appropriate bitboards */
   /* Actually, enemy is a bit misleading, since we're really looking for our */
   /* own pieces */
   own = board_empty;
   set_bitboard(&own, square);
   enemy = board->bbc[player>>7];

   occupied = or_bitboards(board->bbc[0], board->bbc[1]);
   empty = invert_bitboard(occupied);
   
   /* Check if there is an enemy knight looking at this square */
   if (and_bitboards(knight_moves[square], and_bitboards(enemy, board->bbp[1]))) {
      return 1;
   }
   
   /* Check if there is an enemy king looking at this square */
   if (and_bitboards(king_moves[square], and_bitboards(enemy, board->bbp[5]))) {
      return 1;
   }
   
   /* Pawns */
   pawns = and_bitboards(enemy, board->bbp[0]);
   if (player==WHITE) {
      if (and_bitboards(shift_bitboard_left_down(own), pawns) || and_bitboards(shift_bitboard_right_down(own), pawns))
         return 1;
   } else {
      if (and_bitboards(shift_bitboard_left_up(own), pawns) || and_bitboards(shift_bitboard_right_up(own), pawns))
         return 1;
   }
   
   /* Rook/queen */
   bp = and_bitboards(enemy, or_bitboards(board->bbp[3], board->bbp[4]));
   if (bp) {
      scan_direction(shift_bitboard_row_up);
      scan_direction(shift_bitboard_row_down);
      scan_direction(shift_bitboard_file_right);
      scan_direction(shift_bitboard_file_left);
   }

   /* Bishop/queen */
   bp = and_bitboards(enemy, or_bitboards(board->bbp[2], board->bbp[4]));
   if (bp) {
      scan_direction(shift_bitboard_right_up);
      scan_direction(shift_bitboard_right_down);
      scan_direction(shift_bitboard_left_up);
      scan_direction(shift_bitboard_left_down);
   }
   
   /* No, clean */
   return 0;
}

int check(BOARD *board, int player)
{
   int square;
   
   assert(board);
   
   square = sbr_longlong(and_bitboards(board->bbc[player>>7], board->bbp[5]));
   
   return covers(board, BLACK-player, square);
}

