#include <string.h>
#include "alphabet.h"
#include "evaluate.h"
#include "hashtab.h"
#include "hashgen.h"
#include "game.h"

volatile int abort_evaluation = 0;
int prune_count = 0;

/* qsort() callback for sorting the move capture list during quiescent search */
/* Sort according to the logic that it is advantageous to take the most */
/*  valuable piece with a piece that is worth less. */
static int capture_value_cmp(const void *p1, const void *p2)
{
   MOVE *m1 = (MOVE *)p1;
   MOVE *m2 = (MOVE *)p2;
   int val1, val2;

   val1 = piece_value(get_captured_piece(*m1)) - piece_value(get_move_piece(*m1));
   val2 = piece_value(get_captured_piece(*m2)) - piece_value(get_move_piece(*m2));
   
   /* If val1 > val2, then we should look at move 1 first, as it yields the most */
   return -(val1-val2);
}

static int quiescence_search(GAMESTATE *game, const int alpha, const int beta, const int horizon_dist)
{
   MOVE_GEN_STRUCT movelist;
   MOVE bestmove = null_move;
   MOVE move;
   int hash1, hash2;
   int valid_moves;
   int best_value;
   int value;
   int local_alpha;
   int local_beta;
   int start_alpha;
   int start_beta;
   int lower, upper, h=-10;
   int is_check;

   evaluate_count++;
   
   /* Setup semi-sensible default values (checkmate) */
   best_value = -CHECKMATE+horizon_dist;
   value = -CHECKMATE+horizon_dist;
   lower = local_alpha = alpha;
   upper = local_beta = beta;
   
   /* Look for position in the hashtable */
   hash1 = game->hash1;
   hash2 = game->hash2;
   h = 0;
   if (retrieve_hashtable(hash1, hash2, &lower, &upper, &h, &bestmove)) {
      /* If the horizon distance was no smaller than it is now, we're ok to break off */
      if(h>horizon_dist) {
         if (lower>=beta) {
            return lower;
         }
         if (upper<=alpha) {
            return upper;
         }

         if (lower>alpha)
            local_alpha = lower;
         
         if (upper<beta)
            local_beta  = upper;
      }
   }

   if (horizon_dist==-4) {
      return evaluate(game);
   }

   start_alpha = local_alpha;
   start_beta = local_beta;

   /* We should really limit the number of moves that are generated, rather */
   /*  than generate all of them throw away everything but captures/promotions */

   /* Get a list of valid moves in this position */   
   move_gen_start(&game->board, game->player, &movelist);
   /* Limit ourselves to captures and promotions - if we're not checked */
   is_check = check(&game->board, game->player);
   if (!is_check) {
      movelist.total.nummoves = movelist.capture.nummoves + movelist.promotion.nummoves;
      qsort(movelist.total.move, movelist.total.nummoves, sizeof(MOVE), capture_value_cmp);
   }

   /* Begin looking at each move in turn - if there are any, ofcourse */
   valid_moves = movelist.total.nummoves;
   
   /* Return if there are no capture or promotion moves at all */
   /* Here we should check for stalemate */
   if (!valid_moves) {
      return evaluate(game);
   }

   /* Try a null-move first - provided we're not being checked or in an */
   /*  ending where we might be under zugzwang */
   if (!is_check && !(game->material[game->player>>7]+game->material[1^game->player>>7]<ENDGAME_MATERIAL)) {
      best_value = value = evaluate(game);
      /* If the null-move generates a cut-off, then break */
      if (value>=beta) {
         return value;
      }
   }

   /* Check if any of the successors will cause an alpha-beta cutoff */
   /* If so, then we won't need to bother with this node at all. */
   while (get_next_move(&movelist, &move)) {
      playmove(game, move);
      if (retrieve_hashtable(hash1, hash2, &lower, &upper, &h, &move) && h>=horizon_dist) {
         takeback(game);
         if(h>horizon_dist) {
            if (lower>=-alpha) {
               store_hashtable(hash1, hash2, -upper, -lower, horizon_dist, move);
               prune_count++;
               return beta;
            }
         }
      } else {
         takeback(game);
      }
   }   
   movelist.total.cmove = 0;

   while (get_next_move(&movelist, &move)) {
      playmove(game, move);
      /* Make sure we don't put ourselves in check */
      if (!check(&game->board, game->player^BLACK)) {
         value = -quiescence_search(game, -beta, -local_alpha, horizon_dist-1);
      } else {
         valid_moves--;
      }
      takeback(game);

      if (abort_evaluation)
         return -INFINITY;
      
      if (value>best_value)
         best_value = value;

      if (best_value>=beta) {    /* We already found a better result earlier */
         prune_count++;
         break;
      }
      
      if (best_value>local_alpha)
         local_alpha = best_value;
   }
   
   if (valid_moves == 0) {
      best_value = evaluate(game);
   }

   /* Store position in hashtable, but only if the previous result was */
   /*  obtained at a shallower depth */
   if (h<horizon_dist) {
      /* Update bounds */
      if (best_value<=start_alpha) {
         upper = best_value;
      }
      if (best_value>=start_beta){
         lower = best_value;
      }
      if (best_value>start_alpha && best_value<start_beta)
         lower = upper = best_value;
      store_hashtable(hash1, hash2, lower, upper, horizon_dist, bestmove);
   }
   
   return best_value;
}

int alphabeta(GAMESTATE *game, const int alpha, const int beta, const int horizon_dist, VARIANT *line)
{
   VARIANT this_line;
   MOVE_GEN_STRUCT movelist;
   MOVE bestmove = null_move;
   MOVE move;
   int hash1, hash2;
   int valid_moves;
   int best_value;
   int value;
   int local_alpha;
   int local_beta;
   int start_alpha;
   int start_beta;
   int lower, upper, h;
   int killers, killerpos;
   int c;
   int is_check;

   this_line.movecount = 0;
   evaluate_count++;

   /* Setup semi-sensible default values (checkmate) */
   best_value = -CHECKMATE+horizon_dist;
   value = -CHECKMATE+horizon_dist;
   lower = local_alpha = alpha;
   upper = local_beta = beta;
   
   /* Look for position in the hashtable */
   hash1 = game->hash1;
   hash2 = game->hash2;
   h = 0;
   if (retrieve_hashtable(hash1, hash2, &lower, &upper, &h, &move)) {
      /* If the horizon distance was no smaller than it is now, we're ok to break off */
      if(h>horizon_dist) {
         if (lower>=beta) {
            prune_count++;
            return lower;
         }
         if (upper<=alpha) {
            prune_count++;
            return upper;
         }
         
         if (upper==lower) {
            return upper;
         }

         if (lower>alpha)
            local_alpha = lower;
         
         if (upper<beta)
            local_beta  = upper;
            
         bestmove = move;
      }
   }
   
   /* Abort evaluation */
   if (abort_evaluation)
      return -INFINITY;

   /* Reached search boundary (quiscence search) */
   /* Abort one ply before horizon if we're not being checked. */
   is_check = check(&game->board, game->player);
   if (horizon_dist==0 || (horizon_dist==1 && !is_check)) {
      line->movecount = 0;
      return quiescence_search(game, local_alpha, local_beta, 0);
   }
   
   /* Get a list of valid moves in this position */   
   move_gen_start(&game->board, game->player, &movelist);
   
   /* Place the best move found in a previous search in front, if there is */
   /*  any, followed by killer moves. */
   killerpos = 0;
   killers = 0;
   if (!moves_are_equal(bestmove, null_move)) {
      for (c=0; c<movelist.total.nummoves;c++) {
         if (moves_are_equal(bestmove, movelist.total.move[c])) {
            move = movelist.total.move[0];
            movelist.total.move[0] = movelist.total.move[c];
            movelist.total.move[c] = move;
            break;
         }
      }
      killerpos = 1;
      if (is_killer(game, movelist.total.move[0], horizon_dist))
         killers=1;
   }

   for (c=killerpos; c<movelist.total.nummoves && killers<2; c++) {
      if (is_killer(game, movelist.total.move[c], horizon_dist)) {
         move = movelist.total.move[killers];
         movelist.total.move[killers] = movelist.total.move[c];
         movelist.total.move[c] = move;
         killers++;
      }
   }
   
   /* Check if any of the successors will cause an alpha-beta cutoff */
   /* If so, then we won't need to bother with this node at all. */
   while (get_next_move(&movelist, &move)) {
      playmove(game, move);
      if (retrieve_hashtable(hash1, hash2, &lower, &upper, &h, &move) && h>=horizon_dist) {
         takeback(game);
         if(h>horizon_dist) {
            if (lower>=-alpha) {
               prune_count++;
               store_hashtable(hash1, hash2, -upper, -lower, horizon_dist, move);
               return beta;
            }
         }
      } else {
         takeback(game);
      }
   }   
   movelist.total.cmove = 0;

   start_alpha = local_alpha;
   start_beta = local_beta;

   /* Begin looking at each move in turn - if there are any, ofcourse */
   valid_moves = movelist.total.nummoves;

   while (get_next_move(&movelist, &move)) {
      playmove(game, move);
      /* Make sure we don't put ourselves in check */
      if (!check(&game->board, game->player^BLACK)) {
         value = -alphabeta(game, -local_beta, -local_alpha, horizon_dist-1, &this_line);
      } else {
         value = -CHECKMATE+horizon_dist;
         valid_moves--;
      }
      takeback(game);

      if (abort_evaluation)
         return -INFINITY;
      
      if (value>best_value) {
         bestmove = move;
         best_value = value;
      }

      if (best_value>=local_beta) {    /* We already found a better result earlier */
         /* Store killer move for future reference */
         store_killer(game, move, horizon_dist);
         prune_count++;
         break;
      }
      
      if (best_value>local_alpha) {
         local_alpha = best_value;
         /* Store main variant */
         line->move[0] = move;
         memcpy(line->move+1, this_line.move, this_line.movecount * sizeof(MOVE));
         line->movecount = this_line.movecount + 1;
      }
   }
   
   /* If there are no legal moves, we either have a mate or a stalemate */      
   if (!valid_moves) {
      if (is_check)
         best_value = -CHECKMATE+horizon_dist;         /* Checkmate */
      else
         best_value = -STALEMATE+horizon_dist;         /* Stalemate */
   }
   
   /* Store position in hashtable, but only if the previous result was */
   /*  obtained at a shallower depth */
   if (h<=horizon_dist) {
      /* Update bounds */
      if (best_value<=start_alpha) {
         upper = best_value;
      }
      if (best_value>=start_beta){
         lower = best_value;
      }
      if (best_value>start_alpha && best_value<start_beta)
         lower = upper = best_value;
      store_hashtable(hash1, hash2, lower, upper, horizon_dist, bestmove);
   }
   
   return best_value;
}
